//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/Powerfield.h
//! @brief     Defines and implements templated class Powerfield.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_DATA_POWERFIELD_H
#define BORNAGAIN_DEVICE_DATA_POWERFIELD_H

#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Frame.h"
#include "Base/Py/PyObject.h"
#include "Base/Types/OwningVector.h"
#include "Base/Util/Assert.h"
#include "Base/Util/ThreadInfo.h"
#include "Device/Data/LLData.h"
#include "Device/Data/PowerfieldIterator.h"
#include <sstream>

using std::size_t;

//! Templated class to store data in multi-dimensional space.
//! Used with type T = double, CumulativeValue, bool
//! @ingroup tools

template <class T>
class Powerfield : public Frame {
public:
    using value_type = T;

    Powerfield();
    Powerfield(const IAxis& xAxis);
    Powerfield(const IAxis& xAxis, const IAxis& yAxis);
    //! Constructor that takes ownership of supplied axes.
    Powerfield(const std::vector<IAxis*>& axes);
    Powerfield(const Powerfield&) = delete;
    Powerfield(Powerfield&&);
    const Powerfield& operator=(const Powerfield&) = delete;
    ~Powerfield() override;
    Powerfield* clone() const;

    void copyFrom(const Powerfield<T>& other);
    Powerfield<double>* meanValues() const;

    // retrieve basic info

    //! Returns total size of data buffer (product of bin number in every dimension).
    size_t allocatedSize() const
    {
        if (m_ll_data)
            return m_ll_data->getTotalSize();
        return 0;
    }

    //! Returns copy of raw data vector
    std::vector<T> getRawDataVector() const;

    //! Returns sum of all values in the data structure
    T totalSum() const;

    // iterators

    friend class PowerfieldIterator<T, Powerfield<T>>;
    friend class PowerfieldIterator<const T, const Powerfield<T>>;

    //! Read/write iterator type
    using iterator = PowerfieldIterator<T, Powerfield<T>>;

    //! Read-only iterator type
    using const_iterator = PowerfieldIterator<const T, const Powerfield<T>>;

    //! Returns  read/write iterator that points to the first element
    iterator begin();

    //! Returns  read-only iterator that points to the first element
    const_iterator cbegin() const;

    //! Returns  read/write iterator that points to the one past last element
    iterator end() { return iterator(this, allocatedSize()); }

    //! Returns  read-only iterator that points to the one past last element
    const_iterator cend() const { return const_iterator(this, allocatedSize()); }

    // coordinate and index functions

    // modifiers

    //! Sets object into initial state (no dimensions, data)
    void clear();

    //! Sets content of output data to specific value
    void setAllTo(const T& value);

    //! Sets new values to raw data vector
    void setRawDataVector(const std::vector<T>& data_vector);

    //! Sets new values to raw data array
    void setRawDataArray(const T* source);

    //! addition-assignment operator for two output data
    const Powerfield<T>& operator+=(const Powerfield<T>& other);

    //! substraction-assignment operator for two output data
    const Powerfield<T>& operator-=(const Powerfield<T>& other);

    //! division-assignment operator for two output data
    const Powerfield<T>& operator/=(const Powerfield<T>& other);

    //! multiplication-assignment operator for two output data
    const Powerfield<T>& operator*=(const Powerfield<T>& other);

    //! Returns value or summed value, depending on T
    double getValue(size_t index) const;

    //! indexed accessor
    T& operator[](size_t index)
    {
        ASSERT(m_ll_data);
        return (*m_ll_data)[index];
    }

    //! indexed accessor (const)
    const T& operator[](size_t index) const
    {
        ASSERT(m_ll_data);
        return (*m_ll_data)[index];
    }

    // helpers

    //! Returns true if object have same dimensions and number of axes bins
    template <class U>
    bool hasSameSizes(const Powerfield<U>& other) const;

    //! Returns true if objects a) have same dimensions b) bin boundaries of axes coincide
    template <class U>
    bool hasSameShape(const Powerfield<U>& other) const;

#ifdef BORNAGAIN_PYTHON
    //! Returns data as Python numpy array
    PyObject* getArray() const;
#endif

    //! Returns true if object is correctly initialized
    bool isInitialized() const;

    //! Allocates memory for current dimensions configuration
    void allocate();

    const Powerfield<double>& normalizedToMaximum();

private:
    LLData<T>* m_ll_data{nullptr};
};

#ifndef USER_API

// --------------------------------------------------------------------- //
// Implementation
// --------------------------------------------------------------------- //

template <class T>
Powerfield<T>::Powerfield()
    : m_ll_data(nullptr)
{
    allocate();
}

template <class T>
Powerfield<T>::Powerfield(const std::vector<IAxis*>& axes)
    : Frame(axes)
{
    for (const IAxis* ax : axes)
        ASSERT(ax->size() > 0);
    allocate();
}

template <class T>
Powerfield<T>::Powerfield(const IAxis& xAxis)
    : Frame({xAxis.clone()})
{
    ASSERT(xAxis.size() > 0);
    allocate();
}

template <class T>
Powerfield<T>::Powerfield(const IAxis& xAxis, const IAxis& yAxis)
    : Frame({xAxis.clone(), yAxis.clone()})
{
    ASSERT(xAxis.size() > 0);
    ASSERT(yAxis.size() > 0);
    ASSERT(xAxis.axisName() != yAxis.axisName());
    allocate();
}

template <class T>
Powerfield<T>::Powerfield(Powerfield<T>&& other)
    : Frame(other)
{
    m_ll_data = other.m_ll_data;

    other.m_ll_data = nullptr;
    other.allocate();
}

template <class T>
Powerfield<T>::~Powerfield()
{
    clear();
    delete m_ll_data;
}

template <class T>
Powerfield<T>* Powerfield<T>::clone() const
{
    auto* result = new Powerfield<T>();
    result->m_axes = m_axes;
    (*result->m_ll_data) = *m_ll_data;
    return result;
}

template <class T>
void Powerfield<T>::copyFrom(const Powerfield<T>& other)
{
    clear();
    m_axes = other.m_axes;
    delete m_ll_data;
    m_ll_data = 0;
    if (other.m_ll_data)
        m_ll_data = new LLData<T>(*other.m_ll_data);
}

template <class T>
Powerfield<double>* Powerfield<T>::meanValues() const
{
    auto* result = new Powerfield<double>(cloned_axes());
    result->allocate();
    for (size_t i = 0; i < m_ll_data->getTotalSize(); ++i)
        (*result)[i] = getValue(i);
    return result;
}

template <class T>
inline std::vector<T> Powerfield<T>::getRawDataVector() const
{
    ASSERT(m_ll_data);
    std::vector<T> result;
    for (size_t i = 0; i < allocatedSize(); ++i)
        result.push_back((*m_ll_data)[i]);
    return result;
}

template <class T>
typename Powerfield<T>::iterator Powerfield<T>::begin()
{
    typename Powerfield<T>::iterator result(this);
    return result;
}

template <class T>
typename Powerfield<T>::const_iterator Powerfield<T>::cbegin() const
{
    typename Powerfield<T>::const_iterator result(this);
    return result;
}

template <class T>
inline T Powerfield<T>::totalSum() const
{
    ASSERT(m_ll_data);
    return m_ll_data->getTotalSum();
}

template <class T>
void Powerfield<T>::clear()
{
    m_axes.clear();
    allocate();
}

template <class T>
void Powerfield<T>::setAllTo(const T& value)
{
    ASSERT(m_ll_data);
    m_ll_data->setAll(value);
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator+=(const Powerfield<T>& other)
{
    ASSERT(m_ll_data);
    *this->m_ll_data += *other.m_ll_data;
    return *this;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator-=(const Powerfield<T>& other)
{
    ASSERT(m_ll_data);
    *this->m_ll_data -= *other.m_ll_data;
    return *this;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator*=(const Powerfield<T>& other)
{
    ASSERT(m_ll_data);
    *this->m_ll_data *= *other.m_ll_data;
    return *this;
}

template <class T>
bool Powerfield<T>::isInitialized() const
{
    if (!m_ll_data)
        return false;
    if (rank() != m_ll_data->rank())
        return false;
    if (!rank())
        return false;
    return true;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator/=(const Powerfield<T>& other)
{
    ASSERT(m_ll_data);
    *this->m_ll_data /= *other.m_ll_data;
    return *this;
}

template <class T>
void Powerfield<T>::allocate()
{
    delete m_ll_data;
    size_t rank = m_axes.size();
    auto* dims = new int[rank];
    for (size_t i = 0; i < rank; ++i)
        dims[i] = (int)axis(i).size();
    m_ll_data = new LLData<T>(rank, dims);
    T default_value = {};
    m_ll_data->setAll(default_value);
    delete[] dims;
}

template <class T>
inline void Powerfield<T>::setRawDataVector(const std::vector<T>& data_vector)
{
    ASSERT(data_vector.size() == allocatedSize());
    for (size_t i = 0; i < allocatedSize(); ++i)
        (*m_ll_data)[i] = data_vector[i];
}

template <class T>
inline void Powerfield<T>::setRawDataArray(const T* source)
{
    for (size_t i = 0; i < allocatedSize(); ++i)
        (*m_ll_data)[i] = source[i];
}

//! Returns true if object have same dimensions
template <class T>
template <class U>
inline bool Powerfield<T>::hasSameSizes(const Powerfield<U>& other) const
{
    if (!isInitialized())
        return false;
    if (!other.isInitialized())
        return false;
    if (rank() != other.rank())
        return false;
    for (size_t i_axis = 0; i_axis < rank(); ++i_axis)
        if (axis(i_axis).size() != other.axis(i_axis).size())
            return false;
    return true;
}

//! Returns true if object have same dimensions and shape of axis
template <class T>
template <class U>
bool Powerfield<T>::hasSameShape(const Powerfield<U>& other) const
{
    if (!hasSameSizes(other))
        return false;

    for (size_t i = 0; i < m_axes.size(); ++i)
        if (axis(i) != other.axis(i))
            return false;
    return true;
}

#ifdef BORNAGAIN_PYTHON
//! Returns data as Python numpy array
template <>
PyObject* Powerfield<double>::getArray() const;
#endif

#endif // USER_API
#endif // BORNAGAIN_DEVICE_DATA_POWERFIELD_H
