//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/DoubleDescriptor.cpp
//! @brief     Implements class DoubleDescriptor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/BaseItem/SessionItem.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Util/Path.h"
#include <utility>

DoubleDescriptor::DoubleDescriptor(QString label_, QString tooltip_, int decimals_,
                                   const RealLimits& limits_, function<void(double)> setter,
                                   function<double()> getter, const variant<QString, Unit>& unit_)
    : label(std::move(label_))
    , tooltip(std::move(tooltip_))
    , decimals(decimals_)
    , limits(limits_)
    , set(std::move(setter))
    , get(std::move(getter))
    , unit(unit_)
{
    // #baLayerEditor the following path creation has to be implemented when SessionItem migration
    // has finished
    path = [] { return QString(); };
}

DoubleDescriptor::DoubleDescriptor(const QString& label, const QString& tooltip,
                                   function<void(double)> setter, function<double()> getter,
                                   const variant<QString, Unit>& unit)
    : DoubleDescriptor(label, tooltip, 3, RealLimits::nonnegative(), setter, getter, unit)
{
}

DoubleDescriptor::DoubleDescriptor(const QString& label, const QString& tooltip, double* var,
                                   const variant<QString, Unit>& unit)
    : DoubleDescriptor(label, tooltip, 3, RealLimits::nonnegative(), var, unit)

{
}

DoubleDescriptor::DoubleDescriptor(const QString& label, const QString& tooltip, int decimals,
                                   const RealLimits& limits, double* var,
                                   const variant<QString, Unit>& unit)
    : DoubleDescriptor(
        label, tooltip, decimals, limits, [=](double v) { *var = v; }, [=] { return *var; }, unit)
{
}

DoubleDescriptor::DoubleDescriptor(const QString& label, const double* var,
                                   const variant<QString, Unit>& unit)
    : DoubleDescriptor(
        label, "", 3, RealLimits::nonnegative(), [=](double v) { *const_cast<double*>(var) = v; },
        [=] { return *var; }, unit)
{
}

DoubleDescriptor::DoubleDescriptor(const QString& label, SessionItem* item,
                                   const variant<QString, Unit>& unit)
    : DoubleDescriptor(
        label, item->toolTip(), item->decimals(), item->limits(),
        [=](double d) { item->setValue(d); }, [=]() { return item->value().toDouble(); }, unit)
{
    path = [=] {
        ASSERT(item->model()); // if assert, item is not completely initialized
        return GUI::Util::Path::getPathFromIndex(item->model()->indexOfItem(item));
    };
}

DoubleDescriptor::DoubleDescriptor(SessionItem* item, const variant<QString, Unit>& unit)
    : DoubleDescriptor(item->displayName(), item, unit)
{
}

DoubleDescriptor::operator double() const
{
    return get();
}
