//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleCoreShellItem.cpp
//! @brief     Implements class ParticleCoreShellItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Sample/MaterialItems.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Support/Type/VariantUtil.h"
#include "GUI/Util/Error.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"
#include "Sample/Scattering/Rotations.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace

ParticleCoreShellItem::ParticleCoreShellItem(const MaterialItems* materials)
    : ItemWithParticles(abundance_tooltip, position_tooltip)
    , m_materials(materials)
{
}

void ParticleCoreShellItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, m_abundance);
    Serialize::rwProperty(s, m_position);
    Serialize::rwOptional(s, "core", m_core, m_materials);
    Serialize::rwOptional(s, "shell", m_shell, m_materials);
}

std::unique_ptr<ParticleCoreShell> ParticleCoreShellItem::createParticleCoreShell() const
{
    std::unique_ptr<Particle> P_core{};
    std::unique_ptr<Particle> P_shell{};
    if (m_core)
        P_core = m_core->createParticle();
    if (m_shell)
        P_shell = m_shell->createParticle();
    if (!P_core || !P_shell)
        throw Error("ParticleCoreShellItem::createParticleCoreShell -> Error. Either "
                    "core or shell particle is undefined.");
    auto P_coreshell = std::make_unique<ParticleCoreShell>(*P_shell, *P_core);
    P_coreshell->setAbundance(abundance());
    P_coreshell->setParticlePosition(position());
    if (auto r = createRotation(); r && !r->isIdentity())
        P_coreshell->setRotation(*r);

    return P_coreshell;
}

ParticleItem* ParticleCoreShellItem::core() const
{
    return m_core.get();
}

void ParticleCoreShellItem::clearCore()
{
    m_core.reset();
}

ParticleItem* ParticleCoreShellItem::createCore(const MaterialItems* materials)
{
    m_core.reset(new ParticleItem(materials));
    m_core->setMaterial(materials->defaultMaterial());
    return m_core.get();
}

ParticleItem* ParticleCoreShellItem::shell() const
{
    return m_shell.get();
}

void ParticleCoreShellItem::clearShell()
{
    m_shell.reset();
}

ParticleItem* ParticleCoreShellItem::createShell(const MaterialItems* materials)
{
    m_shell.reset(new ParticleItem(materials));
    m_shell->setMaterial(materials->defaultMaterial());

    // position is not used for shell item
    m_shell->setPosition({});
    return m_shell.get();
}

QVector<ItemWithParticles*> ParticleCoreShellItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    if (core())
        result << core() << core()->containedItemsWithParticles();
    if (shell())
        result << shell() << shell()->containedItemsWithParticles();
    return result;
}
