//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/PlotUtil/IntensityDataPropertyWidget.cpp
//! @brief     Implements class IntensityDataPropertyWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/PlotUtil/IntensityDataPropertyWidget.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/View/Common/DoubleSpinBox.h"
#include "GUI/View/Common/SessionItemWidget.h"
#include "GUI/View/Tool/LayoutUtils.h"
#include <QAction>
#include <QCheckBox>
#include <QComboBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QLineEdit>

IntensityDataPropertyWidget::IntensityDataPropertyWidget(QWidget* parent)
    : QWidget(parent)
    , parent(parent)
    , m_item(nullptr)
{
    setWindowTitle("Properties");
    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    m_mainLayout = new QFormLayout(this);
    m_mainLayout->setContentsMargins(8, 8, 8, 8);
    m_mainLayout->setSpacing(5);
}

IntensityDataPropertyWidget::~IntensityDataPropertyWidget()
{
    if (m_item)
        m_item->mapper()->unsubscribe(this);
}

void IntensityDataPropertyWidget::setItem(IntensityDataItem* mainItem, IntensityDataItem* realItem)
{
    if (m_item)
        m_item->mapper()->unsubscribe(this);

    GUI::Util::Layout::clearLayout(m_mainLayout);
    m_updaters.clear();
    m_item = mainItem;

    if (!m_item)
        return;

    m_mainLayout->addRow("Axes units:", createComboBox(m_item->axesUnits()));
    m_mainLayout->addRow("Color scheme:", createComboBox(m_item->gradient()));

    m_mainLayout->addRow(createCheckBox(
        "Interpolate", [=]() { return m_item->isInterpolated(); },
        [=](bool b) { m_item->setInterpolated(b); }));

    // -- x-axis
    auto* xGroup = new QGroupBox("X axis", this);
    auto* xFormLayout = new QFormLayout(xGroup);
    xFormLayout->setContentsMargins(0, 0, 0, 0);
    xFormLayout->setSpacing(5);

    xFormLayout->addRow("Min:", createDoubleSpinbox(m_item->xAxisItem()->min()));
    xFormLayout->addRow("Max:", createDoubleSpinbox(m_item->xAxisItem()->max()));
    xFormLayout->addRow("Title:", createTextEdit(m_item->xAxisItem()->titleItem()));

    m_mainLayout->addRow(xGroup);

    // -- y-axis
    auto* yGroup = new QGroupBox("Y axis", this);
    auto* yFormLayout = new QFormLayout(yGroup);
    yFormLayout->setContentsMargins(0, 0, 0, 0);
    yFormLayout->setSpacing(5);

    yFormLayout->addRow("Min:", createDoubleSpinbox(m_item->yAxisItem()->min()));
    yFormLayout->addRow("Max:", createDoubleSpinbox(m_item->yAxisItem()->max()));
    yFormLayout->addRow("Title:", createTextEdit(m_item->yAxisItem()->titleItem()));

    m_mainLayout->addRow(yGroup);

    // -- color-axis
    auto* zGroup = new QGroupBox("Color legend", this);
    auto* zFormLayout = new QFormLayout(zGroup);
    zFormLayout->setContentsMargins(0, 0, 0, 0);
    zFormLayout->setSpacing(5);

    zFormLayout->addRow("Min:", createDoubleSpinbox(m_item->zAxisItem()->min()));
    zFormLayout->addRow("Max:", createDoubleSpinbox(m_item->zAxisItem()->max()));
    zFormLayout->addRow(createCheckBox("log10", m_item->zAxisItem()->logScaleItem()));
    zFormLayout->addRow(createCheckBox("Visible", m_item->zAxisItem()->visibilityItem()));

    m_mainLayout->addRow(zGroup);

    updateUIValues();

    // react on external changes (e.g. zooming in customplot shall update the axis values)
    m_item->mapper()->setOnChildPropertyChange(
        [=](SessionItem*, const QString&) { updateUIValues(); }, this);

    // update coordinates on axes units change
    m_item->mapper()->setOnPropertyChange(
        [this](SessionItem* item, const QString& name) {
            DataItem::updateAxesUnits(item, name, jobItem()->instrumentItem());
        },
        this);
    if (realItem)
        realItem->mapper()->setOnPropertyChange(
            [this](SessionItem* item, const QString& name) {
                DataItem::updateAxesUnits(item, name, jobItem()->instrumentItem());
            },
            this);

    m_item->mapper()->setOnItemDestroy([this](SessionItem*) { m_item = nullptr; }, this);
}

const JobItem* IntensityDataPropertyWidget::jobItem() const
{
    SessionItemWidget* sessionItemWidget = dynamic_cast<SessionItemWidget*>(parent);
    if (!sessionItemWidget)
        return nullptr;

    return dynamic_cast<JobItem*>(sessionItemWidget->currentItem());
}

QWidget* IntensityDataPropertyWidget::createDoubleSpinbox(DoubleDescriptor d)
{
    auto* spinBox = new DoubleSpinBox(this, d);
    spinBox->setToolTip(d.tooltip);
    QObject::connect(spinBox, &DoubleSpinBox::baseValueChanged,
                     [=](double newValue) { d.set(newValue); });

    m_updaters << [=]() { spinBox->updateValue(); };

    return spinBox;
}

QWidget* IntensityDataPropertyWidget::createTextEdit(SessionItem* item)
{
    auto* edit = new QLineEdit(this);
    connect(edit, &QLineEdit::textEdited, [=](const QString& t) { item->setValue(t); });

    m_updaters << [=]() { edit->setText(item->value().toString()); };

    return edit;
}

QWidget* IntensityDataPropertyWidget::createCheckBox(const QString& title, SessionItem* item)
{
    return createCheckBox(
        title, [=]() { return item->value().toBool(); }, [=](bool b) { item->setValue(b); });
}

QWidget* IntensityDataPropertyWidget::createCheckBox(const QString& title, function<bool()> getter,
                                                     function<void(bool)> setter)
{
    auto* checkBox = new QCheckBox(title, this);
    connect(checkBox, &QCheckBox::stateChanged, [=]() { setter(checkBox->isChecked()); });

    m_updaters << [=]() {
        QSignalBlocker b(checkBox);
        checkBox->setChecked(getter());
    };

    return checkBox;
}

QWidget* IntensityDataPropertyWidget::createComboBox(SelectionDescriptor<QString> d)
{
    auto* combo = new QComboBox(this);
    combo->addItems(d.options);
    combo->setMaxCount(d.options.size());

    connect(combo, qOverload<int>(&QComboBox::currentIndexChanged),
            [=](int newIndex) { d.setCurrentIndex(newIndex); });

    m_updaters << [=]() {
        QSignalBlocker b(combo);
        combo->setCurrentIndex(d.currentIndex());
    };

    return combo;
}

void IntensityDataPropertyWidget::updateUIValues()
{
    for (const auto& updater : m_updaters)
        updater();
}
