//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Tool/WidgetUtils.cpp
//! @brief     Implements GUI::Util namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Tool/WidgetUtils.h"
#include "GUI/Model/Descriptor/UIntDescriptor.h"
#include "GUI/View/Common/DoubleSpinBox.h"
#include "GUI/View/Common/ScientificSpinBox.h"
#include "GUI/View/Tool/EditUtil.h"
#include <QFormLayout>
#include <QLabel>
#include <QSpinBox>

QSpinBox* GUI::Util::createSpinBox(QWidget* parent, const UIntDescriptor& d,
                                   std::function<void(uint)> slot)
{
    auto* spinBox = new QSpinBox(parent);
    spinBox->setFocusPolicy(Qt::StrongFocus);
    spinBox->setToolTip(d.tooltip);
    spinBox->setMaximum(std::numeric_limits<int>::max());
    spinBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    if (d.limits.hasLowerLimit())
        spinBox->setMinimum(static_cast<int>(d.limits.lowerLimit()));
    if (d.limits.hasUpperLimit())
        spinBox->setMaximum(static_cast<int>(d.limits.upperLimit()));

    spinBox->setValue(d.get());

    if (slot)
        QObject::connect(spinBox, qOverload<int>(&QSpinBox::valueChanged),
                         [=](int v) { slot(static_cast<uint>(v)); });

    return spinBox;
}

DoubleSpinBox* GUI::Util::createSpinBox(QFormLayout* parentLayout, const DoubleDescriptor& d,
                                        std::function<void(double)> slot)
{
    auto* sb = new DoubleSpinBox(parentLayout->parentWidget(), d);
    parentLayout->addRow(labelWithUnit(d.label, d.unit) + ":", sb);

    if (slot)
        QObject::connect(sb, &DoubleSpinBox::baseValueChanged, [=](int v) { slot(v); });

    return sb;
}

QSpinBox* GUI::Util::createSpinBox(QFormLayout* parentLayout, const UIntDescriptor& d)
{
    auto* sb = createSpinBox(parentLayout->parentWidget(), d);
    parentLayout->addRow(labelWithUnit(d.label, d.unit) + ":", sb);
    return sb;
}

QString GUI::Util::labelWithUnit(const QString& label, variant<QString, Unit> unit)
{
    const QString s = std::holds_alternative<QString>(unit) ? std::get<QString>(unit)
                                                            : unitAsString(std::get<Unit>(unit));

    if (!s.isEmpty())
        return label + " [" + s + "]";

    return label;
}

QString GUI::Util::labelWithUnit(const DoubleDescriptor& d)
{
    return labelWithUnit(d.label, d.unit);
}

ScientificSpinBox* GUI::Util::createScientificSpinBox(QFormLayout* parentLayout,
                                                      const DoubleDescriptor& d,
                                                      std::function<void(double)> slot)
{
    auto* spinBox = new ScientificSpinBox(parentLayout->parentWidget());
    spinBox->setFocusPolicy(Qt::StrongFocus);
    GUI::View::EditUtil::configSpinbox(spinBox, d.decimals, d.limits);
    spinBox->setToolTip(d.tooltip);
    spinBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    spinBox->setValue(d.get());
    parentLayout->addRow(labelWithUnit(d.label, d.unit) + ":", spinBox);

    if (slot)
        QObject::connect(spinBox, &ScientificSpinBox::valueChanged, [=](double v) { slot(v); });

    return spinBox;
}
