//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/TruncatedSphere.cpp
//! @brief     Implements class TruncatedSphere.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/TruncatedSphere.h"
#include "Base/Math/Bessel.h"
#include "Base/Math/Constants.h"
#include "Base/Math/IntegratorGK.h"
#include "Sample/Shapes/TruncatedEllipsoidNet.h"
#include <limits>

TruncatedSphere::TruncatedSphere(const std::vector<double> P)
    : IFormFactor(P)
    , m_radius(m_P[0])
    , m_height(m_P[1])
    , m_dh(m_P[2])
{
    checkNodeArgs();
    if (m_height > 2. * m_radius || m_dh > m_height) {
        std::ostringstream ostr;
        ostr << "TruncatedSphere: -> Error in class initialization ";
        ostr << "with parameters 'radius':" << m_radius << " 'height':" << m_height
             << " 'delta_height':" << m_dh << "\n\n";
        ostr << "Check for height <= 2.*radius AND delta_height < height failed.";
        throw std::runtime_error(ostr.str());
    }
    m_shape3D =
        std::make_unique<TruncatedEllipsoidNet>(m_radius, m_radius, m_radius, m_height, m_dh);
}

TruncatedSphere::TruncatedSphere(double radius, double height, double dh)
    : TruncatedSphere(std::vector<double>{radius, height, dh})
{
}

//! Complex form factor.
complex_t TruncatedSphere::formfactor_at_bottom(C3 q) const
{
    if (std::abs(q.mag()) < std::numeric_limits<double>::epsilon())
        return M_PI / 3.
               * (m_height * m_height * (3. * m_radius - m_height)
                  - m_dh * m_dh * (3. * m_radius - m_dh));

    complex_t integral = ComplexIntegrator().integrate(
        [=](double Z) {
            double Rz = std::sqrt(m_radius * m_radius - Z * Z);
            complex_t qx = q.x();
            complex_t qy = q.y();
            complex_t q_p = std::sqrt(qx * qx + qy * qy); // NOT the modulus!
            return Rz * Rz * Math::Bessel::J1c(q_p * Rz) * exp_I(q.z() * Z);
        },
        m_radius - m_height, m_radius - m_dh);
    return M_TWOPI * integral * exp_I(q.z() * (m_height - m_radius));
}
