//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/DepthProbeSimulation.h
//! @brief     Defines class DepthProbeSimulation
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_DEPTHPROBESIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_DEPTHPROBESIMULATION_H

#include "Base/Element/DepthProbeElement.h"
#include "Device/Data/Powerfield.h"
#include "Sim/Simulation/ISimulation2D.h"

#include <vector>

class IAxis;
class IComputation;
class IFootprintFactor;
class ICoordSystem;

//! Simulation of radiation depth profile.
//!
//! Holds an instrument and sample model.
//! Computes radiation intensity as function of incoming glancing angle and penetration depth.
//! Scattered rays are neglected.
//! Only refraction, reflection and attenuation of the incoming beam are accounted for.
//! @ingroup simulation

class DepthProbeSimulation : public ISimulation2D {
public:
    DepthProbeSimulation(const MultiLayer& sample);
    ~DepthProbeSimulation() override;

    std::string className() const final { return "DepthProbeSimulation"; }

    //! Sets beam parameters with alpha_i of the beam defined in the range.
    void setBeamParameters(double lambda, int nbins, double alpha_i_min, double alpha_i_max,
                           const IFootprintFactor* beam_shape = nullptr);

    //! Set z positions for intensity calculations. Negative z's correspond to the area
    //! under sample surface. The more negative z is, the deeper layer corresponds to it.
    void setZSpan(size_t n_bins, double z_min, double z_max);

    //! Returns a pointer to incident angle axis.
    const IAxis* alphaAxis() const;

    //! Returns a pointer to z-position axis.
    const IAxis* zAxis() const;

    //! Returns the total number of the intensity values in the simulation result
    size_t intensityMapSize() const override;

    bool force_polarized() const override { return false; }

#ifndef SWIG
    ICoordSystem* createCoordSystem() const override;

private:
    void prepareSimulation() override {}

    SimulationResult pack_result() override;

    //! Sets beam parameters with alpha_i of the beam defined in the range.
    void setBeamParameters(double lambda, const IAxis& alpha_axis,
                           const IFootprintFactor* beam_shape);

    //! Initializes the vector of ISimulation elements
    void initElementVector() override;

    //! Gets the number of elements this simulation needs to calculate
    size_t numberOfElements() const override;

    //! Generate simulation elements for given beam
    std::vector<DepthProbeElement> generateElements(const Beam& beam);

    //! Generate a single threaded computation for a given range of simulation elements
    //! @param re_sample Preprocessed version of our sample
    //! @param start Index of the first element to include into computation
    //! @param n_elements Number of elements to process
    std::unique_ptr<IComputation> createComputation(const reSample& re_sample, size_t start,
                                                    size_t n_elements) override;

    //! Checks if simulation data is ready for retrieval.
    void validityCheck() const;

    void checkCache() const;

    //! Checks the distribution validity for simulation.
    void validateParametrization(const ParameterDistribution& par_distr) const override;

    //! Normalize the detector counts to beam intensity, to solid angle, and to exposure angle.
    //! @param start_ind Index of the first element to operate on
    //! @param n_elements Number of elements to process
    void normalize(size_t start_ind, size_t n_elements) override;

    void addBackgroundIntensity(size_t start_ind, size_t n_elements) override;

    void addDataToCache(double weight) override;

    void moveDataFromCache() override;

    double incidentAngle(size_t index) const;

    //! Creates intensity data from simulation elements
    std::unique_ptr<Powerfield<double>> createIntensityData() const;

    std::unique_ptr<IAxis> m_alpha_axis;
    std::unique_ptr<IAxis> m_z_axis;
    std::vector<DepthProbeElement> m_eles;
    std::vector<std::valarray<double>> m_cache;
#endif
};

#endif // BORNAGAIN_SIM_SIMULATION_DEPTHPROBESIMULATION_H
