//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/OffspecSimulation.h
//! @brief     Defines class OffspecSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H

#include "Sim/Simulation/ISimulation2D.h"

class IAxis;

template <class T>
class Powerfield;

//! Off-specular scattering simulation.
//!
//! Holds an instrument and sample model.
//! Computes reflected and scattered intensity as function of incident and final glancing angle.
//! @ingroup simulation

class OffspecSimulation : public ISimulation2D {
public:
    OffspecSimulation(const Beam& beam, const MultiLayer& sample, const IDetector2D& detector);
    OffspecSimulation(const MultiLayer& sample);
    ~OffspecSimulation() override = default;

    std::string className() const final { return "OffspecSimulation"; }

    //! Sets beam parameters from here (forwarded to Instrument)
    void setBeamParameters(double wavelength, const IAxis& alpha_axis, double phi_i);

    //! Returns axis of the beam.
    const IAxis* beamAxis() const;

    //! Returns the total number of the intensity values in the simulation result
    size_t intensityMapSize() const override;

#ifndef SWIG
    ICoordSystem* createCoordSystem() const override;

private:
    SimulationResult pack_result() override;

    //! Put into a clean state for running a simulation
    void prepareSimulation() override;

    //! Initializes the vector of ISimulation elements
    void initElementVector() override;

    //! Checks the distribution validity for simulation.
    void validateParametrization(const ParameterDistribution& par_distr) const override;

    //! Default implementation only adds the detector axes
    void updateIntensityMap() override;

    //! Gets the number of elements this simulation needs to calculate
    size_t numberOfElements() const override;

    //! Normalize, apply detector resolution and transfer detector image corresponding to
    //! alpha_i = m_alpha_i_axis->bin(index)
    void transferDetectorImage(size_t index);

    //! Check correct number of axes
    void checkInitialization() const;

    std::unique_ptr<IAxis> m_alpha_i_axis;
    std::unique_ptr<Powerfield<double>> m_intensity_map;
#endif
};

#endif // BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H
