//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ScatteringSimulation.cpp
//! @brief     Implements interface ISimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/ScatteringSimulation.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Powerfield.h"
#include "Device/Detector/IDetector2D.h"
#include "Device/Histo/SimulationResult.h"
#include "Resample/Element/DiffuseElement.h"

ScatteringSimulation::ScatteringSimulation(const Beam& beam, const MultiLayer& sample,
                                           const IDetector2D& detector)
    : ISimulation2D(beam, sample, detector)
{
}

void ScatteringSimulation::prepareSimulation()
{
    if (detector().rank() != 2)
        throw std::runtime_error("ScatteringSimulation::prepareSimulation() "
                                 "-> Error. The detector was not properly configured.");
    ISimulation2D::prepareSimulation();
}

SimulationResult ScatteringSimulation::pack_result()
{
    auto* const coordsys = createCoordSystem();
    const std::unique_ptr<Powerfield<double>> data(detector().createDetectorIntensity(m_eles));
    return {*data, *coordsys};
}

size_t ScatteringSimulation::intensityMapSize() const
{
    // #baROI this can be heavily improved by using regionOfInterest directly??!
    size_t result = 0;
    detector().iterateOverRegionOfInterest([&result](IDetector::const_iterator) { ++result; });
    return result;
}

#ifndef SWIG
ICoordSystem* ScatteringSimulation::createCoordSystem() const
{
    auto* const det2D = dynamic_cast<const IDetector2D*>(getDetector());
    ASSERT(det2D);
    return det2D->scatteringCoords(beam());
}
#endif

void ScatteringSimulation::initElementVector()
{
    m_eles = generateElements(beam());
    if (m_cache.empty())
        m_cache.resize(m_eles.size(), 0.0);
}
